#!ruby
# Delphi Perser
# author: YOSHIDA Kazuhiro (moriq)
# mailto: moriq@moriq.com

def putch
  @offset-= @chsize
  @chsize = @prechsize
  @ch = @source[@offset, @chsize]
end

def getch
  raise if @source[@offset].nil?
  @prechsize = @chsize
  if @source[@offset] >= 0x80
    @chsize = 2
  else
    @chsize = 1
  end
  @ch = @source[@offset, @chsize]
  @offset+= @chsize
  @ch
end

@cond_stack = [true]
ConditionalDefine = %w/ MSWINDOWS /
@cond_table = {}
ConditionalDefine.each do |i|
  @cond_table[i.downcase] = true
end

def apply_compiler_directive(cd)
  cd.strip!
  cd.downcase!
  case cd
  when /^\$ifdef\s+(\w+)/
    conditional_symbol = $1
    @cond_stack.push @cond_table[conditional_symbol]
  when /^\$ifndef\s+(\w+)/
    conditional_symbol = $1
    @cond_stack.push ! @cond_table[conditional_symbol]
  when /^\$else/
    @cond_stack.push ! @cond_stack.pop
  when /^\$define\s+(\w+)/
    conditional_symbol = $1
    @cond_table[conditional_symbol] = true
  when /^\$undef\s+(\w+)/
    conditional_symbol = $1
    @cond_table[conditional_symbol] = nil
  when /^\$endif/
    @cond_stack.pop
  end
end

def gettoken
  token = ''
  case @ch
  when "'"
    token+= @ch
    while true
      getch
      case @ch
      when "'"
        token+= @ch
        getch
        if @ch != "'"
          break
        end
      when "\n"
        raise
      end
      token+= @ch
    end
  when "#"
    token+= @ch
    while true
      getch
      unless /\d/ =~ @ch
        break
      end
      token+= @ch
    end
    raise if token.empty?
  # token = token.to_i
  when "$"
    token+= @ch
    while true
      getch
      unless /[\dA-Fa-f]/ =~ @ch
        break
      end
      token+= @ch
    end
    raise if token.empty?
  # token = token.hex
  when /\d/
    point_p = false
    while true
      token+= @ch
      getch
      case @ch
      when '.'
        break if point_p
        getch
        case @ch
        when /\./
          putch
          break
        else
          token+= '.'
        end
        point_p = true
      when 'E', 'e'
        break unless point_p
        getch
        case @ch
        when /[\d+-]/
          token+= 'e'
        else
          putch
          break
        end
      when /\d/
        #
      else
        break
      end
    end
  # token = token.to_f
  when "{"
getch
if @ch == "$"
  # Delphi compiler directive
  cd = '$'
    while true
      getch
      cd+= @ch
      if @ch == '}'
        break
      end
    end
    getch
  cd.chop!  # }
  apply_compiler_directive(cd)
else
putch
  cm = ''
    while true
      getch
      cm+= @ch
      if @ch == '}'
        break
      end
    end
    getch
  cm.chop!  # }
  outputs "comment #{cm}"
end
    token = ' '
  when "("
    getch
    if @ch == "*"
getch
if @ch == "$"
  # Delphi compiler directive
  cd = '$'
    state = 0
    while true
      getch
      cd+= @ch
      case state
      when 0
        if @ch == '*'
          state = 1
        end
      when 1
        if @ch == ')'
          break
        end
        state = 0
      end
    end
    getch
  cd.chop!  # )
  cd.chop!  # *
  apply_compiler_directive(cd)
else
putch
  cm = ''
    state = 0
    while true
      getch
      cm+= @ch
      case state
      when 0
        case @ch
        when '*'
          state = 1
        end
      when 1
        case @ch
        when ')'
          break
        when '*'
        else
          state = 0
        end
      end
    end
    getch
  cm.chop!  # )
  cm.chop!  # *
  outputs "comment #{cm}"
end
      token = ' '
    else
      token = "("
    end
  when "/"
    getch
    if @ch == "/"
      begin
        getch
      end until /\n/ =~ @ch
      getch
      token = ' '
    else
      token = "/"
    end
  when "<"
    token = @ch
    getch
    case @ch
    when "=", ">"
      token+= @ch
      getch
    end
  when ">"
    token = @ch
    getch
    case @ch
    when "="
      token+= @ch
      getch
    end
  when ":"
    token = @ch
    getch
    case @ch
    when "="
      token+= @ch
      getch
    end
  when "."
    token = @ch
    getch
    case @ch
    when "."
      token+= @ch
      getch
    end
  when /\s/
    begin
      getch
    end while /\s/ =~ @ch
    token = ' '
  when /\w/
    begin
      token+= @ch#.downcase
      getch
    end while /\w/ =~ @ch
  else
    token = @ch
    getch
  end
  return token, @cond_stack[-1]
end

def next_token
  ret = @token
  if @tmptoken
    @token = @tmptoken
    @tmptoken = nil
  else
    begin
      token, cond = gettoken
    end while token == ' ' || ! cond
    @token = token
  end
outp ret
  ret
end

def put_token(token)
  @tmptoken = @token
  @token = token
end

def load(fname)
  @source = open(fname).read
  @offset = 0
  getch
  next_token
end

Reserved = %w/ unit interface dispinterface implementation begin end uses type const var threadvar procedure function constructor destructor property resourcestring case /
# private protected public published
ReservedCheck = {}
Reserved.each do |i|
  ReservedCheck[i] = true
end

def ident?(str)
  return false if ReservedCheck[str]
  /^['$#\w]/ =~ str
end

def check(i)
  outputs "check #{@token} #{i}"
  ret = @token
  case i
  when :ident
    raise "`#{ret}' is not ident" unless ident?(ret)
  else
    raise "`#{ret}' is not `#{i}'" unless ret == i
  end
  next_token
end

def class_visibility
  case @token
  when 'private', 'protected', 'public', 'published'
    next_token
# else
#   raise
  end
end

def list_loop(delimiter)
  for loop_count in 0...5096
    yield
    case @token
    when delimiter
      next_token
    else
      break
    end
  end
end

def ident_list
  ret = []
  list_loop(',') { ret.push check :ident }
  ret
end

def class_herit
  check '('
  ret = ident_list
  check ')'
  ret
end

module Delphi
  class Field
    attr_accessor :ident_list, :type
  end
end

def field_decl
  ret = Delphi::Field.new
  ret.ident_list = ident_list
  check(':')
  ret.type = normal_type
  ret
end

module Delphi
  class RecVariant
    attr_accessor :expr_list, :type, :field_list
    def initialize
      @expr_list = []
    end
  end
end

def rec_variant
  ret = Delphi::RecVariant.new
  begin
    list_loop(','){
      expr = []
      expr.push check :ident
      while true
        case @token
        when 'or'
          expr.push next_token
          expr.push check :ident
        else
          break
        end
      end
      ret.expr_list.push expr
    }
  rescue
    #
  end
  case @token
  when ':'
    next_token
    check '('
    ret.field_list = field_list
    check ')'
  end
  ret
end

module Delphi
  class Variant
    attr_accessor :ident, :type, :rec_variant_list
    def initialize
      @rec_variant_list = []
    end
  end
end

def variant_section
  ret = Delphi::Variant.new
  check 'case'
  ret.ident = check :ident
  case @token
  when ':'
    next_token
    ret.type = type_ident
  end
  check 'of'
  list_loop(';'){ ret.rec_variant_list.push rec_variant }
  ret
end

module Delphi
  class FieldList
    attr_accessor :field_decl_list, :variant
    def initialize
      @field_decl_list = []
    end
  end
end

def field_list
  outputs "### field_list"
  ret = Delphi::FieldList.new
  begin
    list_loop(';'){ ret.field_decl_list.push field_decl }
  rescue
    #
  end
  case @token
  when 'case'
    ret.variant = variant_section
  end
  ret
end

module Delphi
  class ObjFieldList
    attr_accessor :ident_list, :type
  end
end

def obj_field_list
  outputs "### obj_field_list"
  ret = Delphi::ObjFieldList.new
  list_loop(';') do
    ret.ident_list = ident_list
    check(':')
    ret.type = normal_type
  end
  ret
end

module Delphi
  class Param
    attr_accessor :ident_list, :type, :value
  end
end

def param
  ret = Delphi::Param.new
  ret.ident_list = ident_list
  case @token
  when ':'
    next_token
    ret.type =
    case @token
    when 'array'
      array_type = []
      array_type.push next_token
      array_type.push check 'of'
      array_type.push \
      case @token
      when 'const'
        next_token
      else
        ord_type
      end
      array_type
    when 'string'
      next_token
    when 'file'
      next_token
    else
      normal_type
    end
    case @token
    when '='
      next_token
      ret.value = const_expr
    end
  end
  ret
end

def formal_param
  case @token
  when 'var', 'const', 'out'
    next_token
  end
  param
end

def formal_param_list
  ret = []
  check '('
  begin
    list_loop(';'){ ret.push formal_param }
  rescue
    #
  end
  check ')'
  ret
end

def directive
  outputs "### directive"
  delim = nil
  case @token
  when ';'
    delim = next_token
  end
  ret = []
for loop_count in 0...5096
  case @token
  when 'cdecl', 'register', 'dynamic', 'virtual', 'export', 'near', 'far', 'forward', 'override', 'overload', 'pascal', 'reintroduce', 'safecall', 'stdcall', 'varargs', 'local', 'abstract', 'assembler'
    ret.push next_token
  when 'external'
    ret.push next_token
    const_expr
  else
    break
  end
  case @token
  when ';'
    delim = next_token
  end
end
    put_token(delim) if delim
  ret
end

def portability_directive
  outputs "### portability_directive"
  delim = nil
  case @token
  when ';'
    delim = next_token
  end
  case @token
  when 'deprecated', 'library', 'platform'
    next_token
  else
    put_token(delim) if delim
  end
end

module Delphi
  class Procedure
    attr_accessor :ident, :param_list, :directive
  end
end

def procedure_header
  ret = Delphi::Procedure.new
  check 'procedure'
  ret.ident = check :ident
  case @token
  when '.'
    next_token
    check :ident
    check '='
    check :ident  # assign interface method
    return
  when '('
    ret.param_list = formal_param_list
  end
  delim = nil
  case @token
  when ';'
    delim = next_token
  end
  case @token
  when 'message'
    next_token
    const_expr
  else
    put_token(delim) if delim
  end
  ret.directive = directive
  portability_directive
  ret
end

module Delphi
  class Function
    attr_accessor :ident, :param_list, :type, :directive
  end
end

def function_header
  ret = Delphi::Function.new
  check 'function'
  ret.ident = check :ident
  case @token
  when '.'
    next_token
    check :ident
    check '='
    check :ident  # assign interface method
    return
  when '('
    ret.param_list = formal_param_list
  end
  check ':'
  ret.type = normal_type
  ret.directive = directive
  portability_directive
  ret
end

module Delphi
  class Constructor
    attr_accessor :ident, :param_list, :directive
  end
end

def constructor_header
  ret = Delphi::Constructor.new
  check 'constructor'
  ret.ident = check :ident
  case @token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

module Delphi
  class Destructor
    attr_accessor :ident, :param_list, :directive
  end
end

def destructor_header
  ret = Delphi::Destructor.new
  check 'destructor'
  ret.ident = check :ident
  case @token
  when '('
    ret.param_list = formal_param_list
  end
  ret.directive = directive
  portability_directive
  ret
end

def method_header
  outputs "### method_header"
  case @token
  when 'class'
    next_token
    # class method
  end
  case @token
  when 'procedure'
    procedure_header
  when 'function'
    function_header
  when 'constructor'
    constructor_header
  when 'destructor'
    destructor_header
  end
end

def prop_param_list
  check '['
  list_loop(';') do
    case @token
    when 'var', 'const', 'out'
      next_token
    end
    ident_list
    check ':'
    type_ident
  end
  check ']'
end

def prop_interface
  case @token
  when '['
    prop_param_list
  end
  check ':'
  normal_type
end

module Delphi
  class Property
    attr_accessor :ident, :interface
  end
end

def prop_list
  outputs "### prop_list"
  ret = Delphi::Property.new
  check 'property'
  ret.ident = check :ident
  case @token
  when '[', ':'
    ret.interface = prop_interface
  end
  case @token
  when 'index'
    next_token
    const_expr
  end
  case @token
  when 'read'
    next_token
    type_ident
  end
  case @token
  when 'write'
    next_token
    type_ident
  end
  case @token
  when 'stored'
    next_token
    type_ident
  end
  case @token
  when 'default'
    next_token
    const_expr
  when 'nodefault'
    next_token
  end
  case @token
  when 'implements'
    next_token
    list_loop(',') { check :ident }
  end
  case @token
  when 'readonly', 'writeonly'
    next_token
  end
  case @token
  when 'dispid'
    next_token
    const_expr
  end
  delim = nil
  case @token
  when ';'
    delim = next_token
  end
  case @token
  when 'default'
    next_token
  else
    put_token(delim) if delim
  end
  ret
end

def record_field_list
  outputs "### record_field_list"
  ret = []
  begin
    list_loop(';'){ ret.push field_list }
  rescue
    #
  end
  ret
end

def class_field_list
  outputs "### class_field_list"
  ret = []
  begin
    list_loop(';') do
#     class_visibility
      ret.push obj_field_list
    end
  rescue
    #
  end
  ret
end

def method_list
  outputs "### method_list"
  ret = method_header
  case @token
  when 'dispid'
    next_token
    const_expr
  end
  ret
end

def class_method_list
  outputs "### class_method_list"
  ret = []
  begin
    list_loop(';') do
#     class_visibility
      ret.push method_list
    end
  rescue
    #
  end
  ret
end

def class_prop_list
  outputs "### class_prop_list"
  ret = []
  begin
    list_loop(';') do
#     class_visibility
      ret.push prop_list
    end
  rescue
    #
  end
  ret
end

def record_type
  check 'record'
  ret = record_field_list
  check 'end'
  ret
end

module Delphi
  class ClassAttr
    attr_accessor :visibility, :spec, :value
    def initialize(visibility, spec, value)
      @visibility = visibility.intern
      @spec = spec
      @value = value
    end
  end
  class Class
    attr_reader :attrs
    attr_accessor :visibility, :herit
    def initialize
      @attrs = []
      @visibility = 'published'
    end
    def add_fields(fields)
      fields.each{|i|
        @attrs.push Delphi::ClassAttr.new @visibility, :field, i
      }
    end
    def add_methods(methods)
      methods.each{|i|
        @attrs.push Delphi::ClassAttr.new @visibility, :method, i
      }
    end
    def add_props(props)
      props.each{|i|
        @attrs.push Delphi::ClassAttr.new @visibility, :prop, i
      }
    end
  end
  class ClassRef
    attr_accessor :base
  end
end

def class_type
  check 'class'
  dc = Delphi::Class.new
  case @token
  when ';'
    return dc
  when 'of'
    next_token
    cref = Delphi::ClassRef.new
    cref.base = type_ident
    return cref
  when '('
    dc.herit = class_herit
    case @token
    when ';'
      return dc
    when 'end'
      next_token
      return dc
    end
  end
  outputs "### class_type " << @token
  dc.visibility = class_visibility || 'published'
  for loop_count in 0...5096
    case @token
    when 'private', 'protected', 'public', 'published'
      dc.visibility = class_visibility
    when 'procedure', 'function', 'constructor', 'destructor'
      dc.add_methods class_method_list
    when 'property'
      dc.add_props class_prop_list
    when 'end'
      break
    else
      dc.add_fields class_field_list
    end
  end
  check 'end'
  outputs "### leave class_type " << @token
  dc
end

def interface_herit
  check '('
  ret = ident_list
  check ')'
  ret
end

def interface_type
  check 'interface'
  case @token
  when ';'
    return
  when '('
    interface_herit
    case @token
    when ';'
      return
    when 'end'
      next_token
      return
    end
  end
  case @token
  when '['
    next_token
    check :ident
    check ']'
  end
  for loop_count in 0...5096
    case @token
    when 'private', 'protected', 'public', 'published'
      class_visibility
    when 'procedure', 'function', 'constructor', 'destructor'
      class_method_list
    when 'property'
      class_prop_list
    when 'end'
      break
    end
  end
  check 'end'
end

def dispinterface_type
  check 'dispinterface'
  case @token
  when ';'
    return
  when '('
    interface_herit
    case @token
    when ';'
      return
    when 'end'
      next_token
      return
    end
  end
  case @token
  when '['
    next_token
    check :ident
    check ']'
  end
  for loop_count in 0...5096
    case @token
    when 'private', 'protected', 'public', 'published'
      class_visibility
    when 'procedure', 'function', 'constructor', 'destructor'
      class_method_list
    when 'property'
      class_prop_list
    when 'end'
      break
    end
  end
  check 'end'
end

module Delphi
  class PointerType
    attr_accessor :ref, :type
    def initialize
      @ref = []
    end
  end
end

def pointer_type
  ret = Delphi::PointerType.new
  for loop_count in 0...5096
    case @token
    when '^'  # pointer
      ret.ref.push next_token
    else
      break
    end
  end
  ret.type = check :ident
end

module Delphi
  class OrdType
    attr_accessor :expr
  end
end

# def ord_ident
#   check :ident
# end

module Delphi
  class SubRangeType
    attr_accessor :begin_expr, :end_expr
  end
end

# def subrange_type
#   const_expr
#   check '..'
#   const_expr
# end

def ord_type
  outputs "### ord_type"
  case @token
  when '('
    enum_type
  else
    # ord_ident
    # subrange_type
    ord_expr = const_expr
    case @token
    when '..'
      next_token
      ret = Delphi::SubRangeType.new
      ret.begin_expr = ord_expr
      ret.end_expr = const_expr
    else
      ret = Delphi::OrdType.new
      ret.expr = ord_expr
    end
    ret
  end
end

module Delphi
  class ArrayType
    attr_accessor :range, :type
    def initialize
      @range = []
    end
  end
end

def array_type
  ret = Delphi::ArrayType.new
  for loop_count in 0...5096
    case @token
    when 'array'
      next_token
      range = []
      case @token
      when '['
        next_token
        list_loop(','){ range.push ord_type }
        check ']'
      end
      check 'of'
      ret.range.push range
    end
  end
  ret.type =
  case @token
  when 'const'
    next_token
  else
    ord_type
  end
  ret
end

module Delphi
  class StringType
    attr_accessor :type
  end
end

def string_type
  ret = Delphi::StringType.new
  check 'string'
  case @token
  when '['
    next_token
    ret.type = ord_type
    check ']'
  end
  ret
end

module Delphi
  class EnumTypeItem
    attr_accessor :ident, :value
  end
end

def enum_type_item
  ret = Delphi::EnumTypeItem.new
  ret.ident = check :ident
  case @token
  when '='
    next_token
    ret.value = const_expr
  end
  ret
end

module Delphi
  class EnumType
    attr_accessor :items
    def initialize
      @items = []
    end
  end
end

def enum_type
  ret = Delphi::EnumType.new
  check '('
  list_loop(','){ ret.items.push enum_type_item }
  check ')'
  ret
end

module Delphi
  class FileType
    attr_accessor :type
  end
end

def file_type
  ret = Delphi::FileType.new
  check 'file'
  case @token
  when 'of'
    next_token
    ret.type = type_ident
  end
  ret
end

module Delphi
  class SetType
    attr_accessor :type
  end
end

def set_type
  ret = Delphi::SetType.new
  check 'set'
  check 'of'
  ret.type = ord_type
  ret
end

module Delphi
  class ProcedureType
    attr_accessor :ident, :param_list, :directive
  end
end

def procedure_type
  ret = Delphi::ProcedureType.new
  check 'procedure'
  case @token
  when '('
    ret.param_list = formal_param_list
  end
  ret
end

module Delphi
  class FunctionType
    attr_accessor :ident, :param_list, :type, :directive
  end
end

def function_type
  ret = Delphi::FunctionType.new
  check 'function'
  case @token
  when '('
    ret.param_list = formal_param_list
  end
  check ':'
  ret.type = normal_type
  ret
end

def proc_type
  outputs "### proc_type"
  ret =
  case @token
  when 'procedure'
    procedure_type
  when 'function'
    function_type
  end
  case @token
  when 'of'
    next_token
    check 'object'
  end
  ret.directive = directive
  portability_directive
  ret
end

def normal_type
  outputs "### normal_type"
  ret = 
  case @token
  when 'type'
    next_token
    normal_type
  when 'procedure', 'function'
    proc_type
  when 'class'
    class_type
  when 'interface'
    interface_type
  when 'dispinterface'
    dispinterface_type
  when 'packed'
    next_token
    normal_type
  when 'record'
    record_type
  when 'file'
    file_type
  when 'set'
    set_type
  when '('
    enum_type
  when '^'
    pointer_type
  when 'array'
    array_type
  when 'string'
    string_type
  else
    ord_type
  end
  portability_directive
  ret
end

def type_ident
  outputs "### type_ident"
  ret = []
  list_loop('.'){ ret.push check :ident }
  ret
end

def factor
  if /^['$#]/ =~ @token
    ret = []
    while /^['$#]/ =~ @token
      outputs "### string #{@token}"
      ret.push next_token
    end
    return ret
  end
  case @token
  when '@', 'not', '+', '-'
    next_token
  end
  if ident?(@token)
    ret = check :ident
    case @token
    when '('  # function call, cast
      next_token
      begin
        list_loop(','){ expr }
      rescue
        #
      end
      check(')')
      return
    when '['
      next_token
      begin
        list_loop(','){ expr }
      rescue
        #
      end
      check ']'
    when '.'
      next_token
      check :ident
    end
    case @token
    when '^'
      next_token
    end
    return ret
  end
  case @token
  when '('
    ## '(' Expression ')' #???
    ret = []
    next_token
    list_loop(';'){
      begin
        list_loop(','){ expr }
      rescue
        #
      end
      case @token
      when ':'
        next_token
        expr
      end
    }
    check ')'
    return ret
  when '['
    ## SetConstructor
    ret = []
    next_token
    begin
      list_loop(','){ expr }
    rescue
      #
    end
    check ']'
    return ret
  end
  raise
end

def term
  ret = []
  ret.push factor
  for loop_count in 0...5096
    case @token
    when '*', '/', 'div', 'mod', 'and', 'shl', 'shr', 'as'
      ret.push next_token
    when '+', '-', 'or', 'xor'
      ret.push next_token
    when '='
      outputs "### = in term"
    # next_token
      return ret
    when '=', '<>', '<', '>', '<=', '>=', 'in', 'is'
      ret.push next_token
    else
      return ret
    end
    ret.push factor
  end
  ret
end

def expr
  ret = []
  ret.push term
  case @token
  when '..'
    ret.push next_token
    ret.push term
    return ret
  end
  ret
end

def const_expr
  expr
end

def uses
  check 'uses'
  ret = []
  for loop_count in 0...5096
    ret.push check :ident
    case @token
    when ','
      next_token
    when ';'
      next_token
      break
    else
      raise "uses: unknown #{@token}"
    end
  end
  ret
end

module Delphi
  class Const
    attr_accessor :ident, :type, :value, :directive
  end
end

def const_decl
  ret = Delphi::Const.new
  ret.ident = check :ident
  case @token
  when ':'
    next_token
    ret.type = normal_type
  end
  outputs "### check = in const_decl"
  check '='
  ret.value = const_expr
  portability_directive
  ret
end

def const_decl_section
  check 'const'
  ret = []
  begin
    list_loop(';'){ ret.push const_decl }
  rescue
    #
  end
  ret
end

def resourcestring_decl_section
  check 'resourcestring'
  ret = []
  begin
    list_loop(';'){ ret.push const_decl }
  rescue
    #
  end
  ret
end

module Delphi
  class Type
    attr_accessor :ident, :value, :directive
  end
end

def type_decl
  ret = Delphi::Type.new
  ret.ident = check :ident
  check '='
  ret.value = normal_type
  ret.directive = directive
  portability_directive
  ret
end

def type_decl_section
  check 'type'
  ret = []
  begin
    list_loop(';'){ ret.push type_decl }
  rescue
    #
  end
  ret
end

module Delphi
  class Var
    attr_accessor :ident_list, :type, :directive, :absolute, :value
  end
end

def var_decl
  outputs "### var_decl"
  ret = Delphi::Var.new
  ret.ident_list = ident_list
  check ':'
  ret.type = normal_type
  ret.directive = directive
  case @token
  when 'absolute'
    next_token
    ret.absolute = check :ident
  when '='
    next_token
    outputs "### = in var_decl"
    ret.value = const_expr
  end
  portability_directive
  ret
end

def var_decl_section
  check 'var'
  ret = []
  begin
    list_loop(';'){ ret.push var_decl }
  rescue
    #
  end
  ret
end

def threadvar_decl_section
  check 'threadvar'
  ret = []
  begin
    list_loop(';'){ ret.push var_decl }
  rescue
    #
  end
  ret
end

module Delphi
  class Procedure
    attr_accessor :define
  end
end

def procedure_decl
  outputs "### procedure_decl"
  ret = Delphi::Procedure.new
  ret.define = procedure_header
  check ';'
end

module Delphi
  class Function
    attr_accessor :define
  end
end

def function_decl
  outputs "### function_decl"
  ret = Delphi::Function.new
  ret.define = function_header
  check ';'
end

def procedure_decl_section
  ret = []
  case @token
  when 'procedure'
    ret.push procedure_decl
  when 'function'
    ret.push function_decl
  end
  ret
end

module Delphi
  class Interface
    attr_reader :uses, :types, :consts, :resourcestrings, :vars, :threadvars, :procedures
    def initialize
      @uses = []
      @types = []
      @consts = []
      @resourcestrings = []
      @vars = []
      @threadvars = []
      @procedures = []
    end
  end
end

def interface
  check 'interface'
  di = Delphi::Interface.new
  for loop_count in 0...5096
    case @token
    when 'uses'
      di.uses.concat uses
    when 'type'
      di.types.concat type_decl_section
    when 'const'
      di.consts.concat const_decl_section
    when 'resourcestring'
      di.resourcestrings.concat resourcestring_decl_section
    when 'var'
      di.vars.concat var_decl_section
    when 'threadvar'
      di.threadvars.concat threadvar_decl_section
    when 'procedure', 'function'
      di.procedures.concat procedure_decl_section
    else
      break
    end
  end
  di
end

def implementation
  check 'implementation'
end

module Delphi
  class Unit
    attr_accessor :ident, :interface
  end
end

def unit
begin
  check 'unit'
rescue
  errputs $!.message
  $!.backtrace.each{|i| errputs i}
  return
end
  du = Delphi::Unit.new
  du.ident = check :ident
  portability_directive
  check ';'
  du.interface = interface
  implementation
  du
end

def debputs(*args)
  puts *args
end

def outputs(*args)
  @out.puts *args
end

require 'pp'

def outp(*args)
  defout = $>
  $> = @out
  args.each{|i| pp i }
  $> = defout
end

@err = open('err.txt', 'w')
def errputs(*args)
  @err.puts *args
  STDERR.puts *args
end

Dir.glob('**/*.pas') do |fname|
  logfname = fname.tr('./', '__')<<'.txt'
  @out = open(logfname, 'w')
  errputs "### #{fname}"
  load(fname)
=begin
set_trace_func proc{|event, file, line, id, binding, klass|
  case id.to_s
  when /^get/, /^next_/, /^check/, /puts$/, 'outp'
    break
  end
  case event
  when /^raise/, /^call/#, /^return/
    outputs "#{file}(#{line}): #{id} #{event}"
  end
}
=end
begin
  outp unit
rescue
  errputs "!!! #{logfname}"
  errputs $!.message
# $!.backtrace.each{|i| errputs i}
end
set_trace_func nil
STDOUT.flush
STDERR.flush
@err.flush
@out.close
end

